#!/usr/bin/env python
"""libamgtogeo.py: Library to transform Australian Map Grid (AMG) to Geographic co-ordinates according to Redfearn's formulae.

Based on "amg2geo_ascii_agd66.pl", written by Neil Freeman in May 1996.  Thanks to him for providing the Perl code nessesary to port it to Python (originally for use with libubdadl).  Please note that I've merely ported the program, I unfortunately can't explain the mathematics behind it. :-)

This Python version of the code is Copyright 2005 Michael Farrell.

This library is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation either version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA."""

from math import pi, sin, cos

def AMGToGeo(easting, northing, zone):
	"""amgtogeo(easting, northing, zone): Convert an AMG reference to Geographic co-ordinates.  These are returned in a tuple of (latitude, longitude)."""
	if (zone == 49):
		cm = 111.0
	elif (zone == 50):
		cm = 117.0
	elif (zone == 51):
		cm = 123.0
	elif (zone == 52):
		cm = 129.0
	elif (zone == 53):
		cm = 135.0
	elif (zone == 54):
		cm = 141.0
	elif (zone == 55):
		cm = 147.0
	elif (zone == 56):
		cm = 153.0
	else:
		raise ValueError, "Zone outside acceptable range (49 - 56)!"
	
	# set some constants
	k0 = 0.9996
	a = 6378160.0
	f = 1.0 / 298.25
	
	# compute eccentricity
	ecc = ((2.0*f)-(f**2.0))**0.5

	# strip the false origin off the input AMG coordinates
	e = easting - 500000.0
	n = northing - 10000000.0

	# compute meridian distance equal to northing/scale factor
	m = n / k0

	# compute foot-point latitude series expansion
	nfpl = f/(2.0-f)
	nfpl2 = nfpl**2.0
	nfpl3 = nfpl**3.0
	nfpl4 = nfpl**4.0
	gfpl = a*(1.0-nfpl)*(1.0-nfpl2)*(1.0+((9.0/4.0)*nfpl2)+((255.0/64.0)*nfpl4))*(pi/180.0)
	sig = (m*pi)/(gfpl*180.0)
	fpl = sig+(((3.0*nfpl)/2.0)-((27.0*nfpl3)/32.0))*sin(2.0*sig)+(((21.0*nfpl2)/16.0)-((55.0*nfpl4)/32.0))*sin(4.0*sig)+((151.0*nfpl3)/96.0)*sin(6.0*sig)+((1097.0*nfpl4)/512.0)*sin(8.0*sig)

	# compute radii of curvature and other variables
	p = a*(1.0-(ecc**2.0))/(1.0-((ecc**2.0)*(sin(fpl)**2.0)))**1.5
	v = a/(1.0-((ecc**2.0)*(sin(fpl)**2.0)))**0.5
	u = v/p
	t = sin(fpl)/cos(fpl)

	# compute powers of term elements, tc
	x = e/(k0*v)
	tcop = t/(k0*p)
	x3 = x**3.0
	x5 = x**5.0
	x7 = x**7.0
	t2 = t**2.0
	t4 = t**4.0
	t6 = t**6.0
	u2 = u**2.0
	u3 = u**3.0
	u4 = u**4.0
	sec = 1.0/cos(fpl)

	# compute latitude
	term1 = tcop*x*e/2.0
	term2 = tcop*(x3*e/24.0)*((-4.0*u2)+((9.0*u)*(1.0-t2))+(12.0*t2))
	term3 = tcop*(x5*e/720.0)*(((8.0*u4)*(11.0-(24.0*t2)))-((12.0*u3)*(21.0-(71.0*t2)))+((15.0*u2)*(15.0-(98.0*t2)+(15.0*t4)))+((180.0*u)*((5.0*t2)-(3.0*t4)))+(360.0*t4))
	term4 = tcop*(x7*e/40320.0)*(1385.0+(3633.0*t2)+(4095.0*t4)+(1575.0*t6))
	rlat = fpl - term1 + term2 - term3 + term4
	latitude = rlat*(180.0/pi)

	# compute longitude
	term1 = sec*(x3/6.0)*(u+(2.0*t2))
	term2 = sec*(x5/120.0)*(((-4.0*u3)*(1.0-(6.0*t2)))+(u2*(9.0-(68.0*t2)))+(72.0*u*t2)+(24.0*t4))
	term3 = sec*(x7/5040.0)*(61.0+(662.0*t2)+(1320.0*t4)+(720.0*t6))
	omega = (sec*x) - term1 + term2 - term3
	domega = omega*(180.0/pi)
	longitude = domega + cm

	return (latitude, longitude)
