#!/usr/bin/env python
"""libubdadl: Transform an Adelaide UBD grid coordinate into a AMG grid ref or Geographic ref.
Copyright 2005 Michael Farrell.

This library is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation either version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.

Known limitations:
 - All the maps have to be nicely lined up in a grid.  Those that don't can't be included.  So south of Aldinga bay doesn't work, and has been disabled.
 - This doesn't count the inset maps.
 - Some maps may not exist in your version of the UBD.  Probably because they are outdated, or just the UBD doesn't really cover them yet.  They're unlocked, and you can get co-ordinates for places if that map actually existed.
 - True north and grid north are off by about 1 degree.  This is UBD's problem, however this is insignificant.
 - The Adelaide CBD (maps 1 - 4) don't work.  Don't bother me about this, the scale is different and would probably incur another version of libubdadl just for the CBD.  If accuracy is a problem, buy a GPS rather than using this.  Try to get by with map 118.
 - The regional and major connecting roads maps aren't supported.  They're too inaccurate anyway. (They cover at least three to five times the area).
 - The accuracy is to within a 125 metre square.  This is a limitation of the maps, keeping with the KISS policy of interface.

This library takes the following data:
 - A map number
 - A grid reference
 - A "subgrid" reference

and gives you GPS coordinates for your wardriving software.

A subgrid is the position of the place within the 250 x 250 metre grid.  It's divided into a 3 x 3 grid of points, each cover upto a radius of 83 meters.

 +----+----+
 | NW N NE |
 +  W C E  +
 | SW S SE |
 +----+----+

The exact position of the measurement is at the plus nearest the direction, with the exception of C which is on the C itself.  The points overlap, so E on A1 is W on A2.  All in all, each square really only presents four points, however nine can be addressed.  How you want to do this is all up to user preference.

For more info, see doc/libudbadl-howto.html.
"""

import string, libamgtogeo

### DEFINES ###

def startingPoint():
	"""The starting point on AMG we begin at.  First easting, then northing, then map number."""
	return (268000, 6175000, 54)

def mapRange():
	"""The range of maps we will accept.  Note that not all of these are valid, but this varies between versions of the UBD.  We do not support >214, as those maps are offset for the start of Aldinga Bay."""
	return (7, 218)

def mapNL():
	"""Which maps are map "newlines".  Note that not all of these exist yet."""
	return (6, 16, 26, 36, 46, 56, 66, 78, 90, 102, 114, 126, 138, 150, 162, 174, 184, 194, 204, 214)

def cellSize():
	"""The size of each cell on the map, in meters.  First x then y."""
	return (250, 250)

def mapSize():
	"""The size, in cells, of each map.  First x then y."""
	return (15, 16)

def validLetters():
	"""The X axis can skip letters.  These are the ones that are valid."""
	return ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J', 'K', 'L', 'M', 'N', 'P', 'Q')

### REAL FUNCTIONS ###

def numberToLetter(number):
	"""Translate a number X-value to a letter."""
	letters = validLetters()
	number = number - 1
	return letters[number]

def letterToNumber(letter):
	"""Translate a letter X-value to a number."""
	letters = validLetters()
	letter = string.upper(letter)
	index = 0
	for l in letters:
		if l == letter:
			return index
		index = index + 1
	raise IndexError, "That letter co-ordinate isn't valid."

def gridToNumber(grid):
	"""Convert a two-char grid reference to a zero-indexed coordinate."""
	return (letterToNumber(grid[0]), int(grid[1:])-1)

def mapToXY(mapid):
	"""Create a zero-indexed map cooridinate, with a grid of maps."""
	maprange = mapRange()
	if (mapid < maprange[0]) or (mapid > maprange[1]):
		raise IndexError, "That map is out of range."
	mapnl = mapNL()
	rowstart = 0
	y = -1
	for nl in mapnl:
		if mapid > nl:
			y = y + 1
			rowstart = nl + 1
	x = mapid - rowstart
	return (x, y)

def subgridOffset(cardinal):
	"""Returns an offset for a given cardinal point."""
	cardinal = string.upper(cardinal)
	cellsize = cellSize()
	offset = (0, 0)
	# NW has 0 offset, all the rest are relative to that
	if cardinal == 'N':
		offset = (125, 0)
	elif cardinal == 'NE':
		offset = (250, 0)
	elif cardinal == 'W':
		offset = (0, 125)
	elif cardinal == 'C':
		offset = (125, 125)
	elif cardinal == 'E':
		offset = (250, 125)
	elif cardinal == 'SW':
		offset = (0, 250)
	elif cardinal == 'S':
		offset = (125, 250)
	elif cardinal == 'SE':
		offset = (250, 250)
	return offset

def mapToAMG(mapid, grid, subgrid):
	startingpoint = startingPoint()
	gridXY = gridToNumber(grid)
	mapXY = mapToXY(mapid)
	subgridoff = subgridOffset(subgrid)
	cellsize = cellSize()
	mapsize = mapSize()
	x = startingpoint[0] + (mapsize[0] * cellsize[0] * mapXY[0]) + (cellsize[0] * gridXY[0]) + subgridoff[0]
	y = startingpoint[1] - (mapsize[1] * cellsize[1] * mapXY[1]) - (cellsize[1] * gridXY[1]) - subgridoff[1]
	return (x, y, startingpoint[2])

def mapToGeo(mapid, grid, subgrid):
	amg = mapToAMG(mapid, grid, subgrid)
	geo = libamgtogeo.AMGToGeo(amg[0], amg[1], amg[2])
	return geo
